@@ -32,7 +32,8 @@ import System.FilePath (replaceFileName, takeBaseName, takeDirectory, (</>))
 import Text.Parsec
 import Typst.Bind (destructuringBind)
 import Typst.Methods (getMethod)
-import Typst.Module.Standard (loadFileText, standardModule, symModule)
+import Typst.Module.Standard (loadFileText, standardModule, symModule,
+                              sysModule)
 import Typst.Module.Math (mathModule)
 import Typst.MathClass (mathClassOf, MathClass(Relation))
 import Typst.Parse (parseTypst)
@@ -42,8 +43,7 @@ import Typst.Syntax
 import Typst.Types
 import Typst.Util (makeFunction, nthArg)
 import qualified Toml as Toml
-import qualified Toml.FromValue as Toml
-import qualified Toml.FromValue.Generic as Toml
+import qualified Toml.Schema as Toml
 
 -- import Debug.Trace
 
@@ -57,13 +57,14 @@ evaluateTypst ::
   FilePath ->
   -- | Markup produced by 'parseTypst'
   [Markup] ->
-  m (Either ParseError (Seq Content))
-evaluateTypst operations fp =
+  m (Either ParseError Content)
+evaluateTypst operations =
   runParserT
-    (mconcat <$> many pContent <* eof)
+    (do contents <- mconcat <$> many pContent <* eof
+        -- "All documents are automatically wrapped in a document element."
+        pure $ Elt "document" Nothing [("body", VContent contents)])
     initialEvalState { evalOperations = operations,
-                       evalPackageRoot = takeDirectory fp }
-    fp
+                       evalPackageRoot = "." }
 
 initialEvalState :: EvalState m
 initialEvalState =
@@ -241,6 +242,7 @@ pElt = do
     Equation display ms -> inBlock BlockScope $ do
       importModule mathModule
       importModule symModule
+      importModule sysModule
       oldMath <- evalMath <$> getState
       updateState $ \st -> st {evalMath = True}
       content <- pInnerContents ms
@@ -478,8 +480,12 @@ evalExpr expr =
           ( \m v -> do
               case v of
                 Reg (k, e) -> do
+                  k' <- case k of
+                           Ident i -> pure i
+                           _ -> do VString s <- evalExpr k
+                                   pure $ Identifier s
                   val <- evalExpr e
-                  pure $ m OM.|> (k, val)
+                  pure $ m OM.|> (k', val)
                 Spr y -> do
                   val <- evalExpr y
                   case val of
@@ -630,12 +636,18 @@ evalExpr expr =
             VString t' -> pure $ VBoolean $ t' `T.isInfixOf` t
             VRegex re -> pure $ VBoolean $ match re t
             _ -> fail $ "Can't apply 'in' to " <> show v1 <> " and string"
+        VType ty ->
+          case v1 of
+            VString t' -> pure $ VBoolean $ t' `T.isInfixOf` (prettyType ty)
+            VRegex re -> pure $ VBoolean $ match re (prettyType ty)
+            _ -> fail $ "Can't apply 'in' to " <> show v1
         VArray vec -> pure $ VBoolean $ V.elem v1 vec
         VDict m ->
           case v1 of
             VString t -> pure $ VBoolean $ isJust $ OM.lookup (Identifier t) m
             _ -> pure $ VBoolean False
-        _ -> fail $ "Can't apply 'in' to " <> show v2
+        _ -> fail $ "Can't apply 'in' to " <> show v2 <> show (e1,e2)
+
     Negated e -> do
       v <- evalExpr e
       case maybeNegate v of
@@ -827,13 +839,15 @@ evalExpr expr =
           _ -> fail "Import requires a path or module or function"
       case imports of
         AllIdentifiers -> importModule modmap
-        SomeIdentifiers idents -> do
-          let addFromModule m ident =
+        SomeIdentifiers pairs -> do
+          let addFromModule m (ident, mbAs) =
                 case M.lookup ident modmap of
                   Nothing -> fail $ show ident <> " not defined in module"
-                  Just v -> pure $ M.insert ident v m
-          foldM addFromModule mempty idents >>= importModule
-        NoIdentifiers -> addIdentifier modid (VModule modid modmap)
+                  Just v -> pure $ M.insert (fromMaybe ident mbAs) v m
+          foldM addFromModule mempty pairs >>= importModule
+        NoIdentifiers mbAs -> do
+          let ident = fromMaybe modid mbAs
+          addIdentifier ident (VModule ident modmap)
       pure VNone
     Include e -> do
       argval <- evalExpr e
@@ -915,11 +929,11 @@ data PackageConfig = PackageConfig {
 
 -- | Derived generically from record field names
 instance Toml.FromValue Config where
-  fromValue = Toml.parseTableFromValue Toml.genericParseTable
+  fromValue = Toml.genericFromTable
 
 -- | Derived generically from record field names
 instance Toml.FromValue PackageConfig where
-  fromValue = Toml.parseTableFromValue Toml.genericParseTable
+  fromValue = Toml.genericFromTable
 
 findPackageEntryPoint :: Monad m => Text -> MP m FilePath
 findPackageEntryPoint modname = do
@@ -950,7 +964,7 @@ findPackageEntryPoint modname = do
   let localDir = dataDir </> "typst"
   let cacheDir = cacheDir' </> "typst"
 #endif
-  let subpath = "packages" </> namespace </> (name <> "-" <> version)
+  let subpath = "packages" </> namespace </> name </> version
   inLocal <- lift $ checkExistence operations (localDir </> subpath </> "typst.toml")
   tomlPath <-
      if inLocal
@@ -963,7 +977,7 @@ findPackageEntryPoint modname = do
                     (localDir </> subpath) ++ "\n" ++ (cacheDir </> subpath) ++
                     "\nCompile with typst compile to bring the package into your local cache."
              -- TODO? fetch from CDN if not present in cache?
-  tomlString <- T.unpack . TE.decodeUtf8 <$> lift (loadBytes operations tomlPath)
+  tomlString <- TE.decodeUtf8 <$> lift (loadBytes operations tomlPath)
   case Toml.decode tomlString of
     Toml.Failure e -> fail (unlines ("Failure loading typst.toml" : e))
     Toml.Success _warnings cfg -> -- ignores warnings like unused keys in TOML
@@ -973,17 +987,24 @@ loadModule :: Monad m => Text
            -> MP m (Seq Content, (Identifier, M.Map Identifier Val))
 loadModule modname = do
   pos <- getPosition
-  (fp, mbPackageRoot) <-
+  (fp, modid, mbPackageRoot) <-
         if T.take 1 modname == "@"
            then do
             fp' <- findPackageEntryPoint modname
-            pure (fp', Just (takeDirectory fp'))
+            pure (fp',
+                   Identifier
+                    (T.pack $ takeWhile (/= ':') . takeBaseName $
+                       T.unpack modname),
+                   Just (takeDirectory fp'))
            else if T.take 1 modname == "/" -- refers to path relative to package root
                 then do
                   packageRoot <- evalPackageRoot <$> getState
-                  pure (packageRoot </> drop 1 (T.unpack modname), Nothing)
-                else pure (replaceFileName (sourceName pos) (T.unpack modname), Nothing)
-  let modid = Identifier (T.pack $ takeBaseName fp)
+                  pure (packageRoot </> drop 1 (T.unpack modname),
+                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
+                        Nothing)
+                else pure (replaceFileName (sourceName pos) (T.unpack modname),
+                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
+                        Nothing)
   txt <- loadFileText fp
   case parseTypst fp txt of
     Left err -> fail $ show err
@@ -1114,57 +1135,60 @@ inBlock scope pa = do
   pure result
 
 updateExpression :: Monad m => Expr -> Val -> MP m ()
-updateExpression e val =
-  case e of
-    Ident i -> updateIdentifier i val
-    FuncCall
-      (FieldAccess (Ident (Identifier "at")) e')
-      [NormalArg (Literal (Int idx))] ->
-        do
-          ival <- evalExpr e'
-          case ival of
-            VArray v ->
-              updateExpression e' $ VArray $ v V.// [(fromIntegral idx, val)]
-            _ -> fail $ "Cannot update expression " <> show e
-    FuncCall (FieldAccess (Ident (Identifier "first")) e') [] ->
-      updateExpression
-        ( FuncCall
-            (FieldAccess (Ident (Identifier "at")) e')
-            [NormalArg (Literal (Int 0))]
-        )
-        val
-    FuncCall (FieldAccess (Ident (Identifier "last")) e') [] ->
-      updateExpression
-        ( FuncCall
-            (FieldAccess (Ident (Identifier "at")) e')
-            [NormalArg (Literal (Int (-1)))]
-        )
-        val
-    FuncCall
-      (FieldAccess (Ident (Identifier "at")) e')
-      [NormalArg (Literal (String fld))] ->
-        do
-          ival <- evalExpr e'
-          case ival of
-            VDict d ->
-              updateExpression e' $
-                VDict $
-                  OM.alter
-                    ( \case
-                        Just _ -> Just val
-                        Nothing -> Just val
-                    )
-                    (Identifier fld)
-                    d
-            _ -> fail $ "Cannot update expression " <> show e
-    FieldAccess (Ident (Identifier fld)) e' ->
-      updateExpression
-        ( FuncCall
-            (FieldAccess (Ident (Identifier "at")) e')
-            [NormalArg (Literal (String fld))]
-        )
-        val
-    _ -> fail $ "Cannot update expression " <> show e
+updateExpression
+  (FuncCall (FieldAccess (Ident (Identifier "first")) e') []) val
+  = updateExpression ( FuncCall
+                       (FieldAccess (Ident (Identifier "at")) e')
+                       [NormalArg (Literal (Int 0))]
+                     ) val
+updateExpression
+  (FuncCall (FieldAccess (Ident (Identifier "last")) e') []) val
+  = updateExpression ( FuncCall
+                       (FieldAccess (Ident (Identifier "at")) e')
+                       [NormalArg (Literal (Int (-1)))]
+                     ) val
+updateExpression
+  (FieldAccess (Ident (Identifier fld)) e') val
+  = updateExpression' True e' (Literal (String fld)) val -- see #26
+updateExpression (Ident i) val = updateIdentifier i val
+updateExpression
+  (FuncCall
+      (FieldAccess (Ident (Identifier "at")) e') [NormalArg arg]) val
+  = updateExpression' False e' arg val
+updateExpression e _ = fail $ "Cannot update expression " <> show e
+
+updateExpression' :: Monad m => Bool -> Expr -> Expr -> Val -> MP m ()
+updateExpression' allowNewIndices e arg val = do
+    container <- evalExpr e
+    idx <- evalExpr arg
+    case container of
+      VArray v ->
+        case idx of
+          VInteger i ->
+            let i' = fromIntegral i
+            in case v V.!? i' of
+                 Nothing | not allowNewIndices
+                     -> fail $ "Vector does not contain index " <> show i'
+                 _ -> updateExpression e $ VArray $ v V.// [(i', val)]
+          _ -> fail $ "Cannot index array with " <> show idx
+      VDict d ->
+        case idx of
+          VString fld ->
+            case OM.lookup (Identifier fld) d of
+              Nothing | not allowNewIndices
+                     -> fail $ "Dictionary does not contain key " <> show fld
+              _ -> updateExpression e $
+                           VDict $
+                             OM.alter
+                               ( \case
+                                   Just _ -> Just val
+                                   Nothing -> Just val
+                               )
+                               (Identifier fld)
+                               d
+          _ -> fail $ "Cannot index dictionary with " <> show idx
+      _ -> fail $ "Cannot update expression " <> show e
+
 
 toSelector :: Monad m => Val -> MP m Selector
 toSelector (VSelector s) = pure s
