@@ -7,6 +7,7 @@
 module Typst.Module.Standard
   ( standardModule,
     symModule,
+    sysModule,
     loadFileText,
     applyPureFunction
   )
@@ -15,7 +16,7 @@ where
 import Data.Char (ord, chr)
 import Control.Applicative ((<|>))
 import Control.Monad (mplus, unless)
-import Control.Monad.Reader (lift)
+import Control.Monad.Reader (lift, asks)
 import qualified Data.Aeson as Aeson
 import qualified Data.ByteString.Lazy as BL
 import qualified Data.Csv as Csv
@@ -41,16 +42,18 @@ import Typst.Regex (makeRE)
 import Typst.Symbols (typstSymbols)
 import Typst.Types
 import Typst.Util
+import System.FilePath ((</>))
+import Data.List (genericTake)
 import Data.Time (UTCTime(..))
 import Data.Time.Calendar (fromGregorianValid)
 import Data.Time.Clock (secondsToDiffTime)
-import Data.Digits (mDigits)
 
 standardModule :: M.Map Identifier Val
 standardModule =
   M.fromList $
     [ ("math", VModule "math" mathModule),
       ("sym", VModule "sym" symModule),
+      ("sys", VModule "sys" sysModule),
       ("emoji", VModule "emoji" emojiModule),
       ("calc", VModule "calc" calcModule)
     ]
@@ -66,6 +69,10 @@ standardModule =
       ++ time
       ++ dataLoading
 
+sysModule :: M.Map Identifier Val
+sysModule =
+  M.fromList [ makeElement (Just "sys") "version" [] ]
+
 symModule :: M.Map Identifier Val
 symModule = M.map VSymbol $ makeSymbolMap typstSymbols
 
@@ -90,6 +97,7 @@ textual =
     makeElement Nothing "smallcaps" [("body", One TContent)],
     makeElement Nothing "underline" [("body", One TContent)],
     makeElement Nothing "overline" [("body", One TContent)],
+    makeElement Nothing "highlight" [("body", One TContent)],
     makeElement Nothing "raw" [("text", One TString)],
     makeElement Nothing "smartquote" [],
     makeElement Nothing "lower" [("text", One (TString :|: TContent))],
@@ -127,7 +135,6 @@ layout =
     makeElement Nothing "box" [("body", One TContent)],
     makeElement Nothing "colbreak" [],
     makeElement Nothing "columns" [("count", One TInteger), ("body", One TContent)],
-    makeElement Nothing "grid" [("children", Many TContent)],
     makeElement Nothing "h" [("amount", One (TLength :|: TRatio :|: TFraction))],
     makeElement Nothing "v" [("amount", One (TLength :|: TRatio :|: TFraction))],
     makeElement Nothing "hide" [("body", One TContent)],
@@ -164,7 +171,24 @@ layout =
       Nothing
       "stack"
       [("children", Many (TLength :|: TRatio :|: TFraction :|: TContent))],
-    makeElement Nothing "table" [("children", Many TContent)],
+    makeElementWithScope Nothing
+      "table"
+      [("children", Many TContent)]
+      [ makeElement (Just "table") "cell" [ ("body", One TContent) ]
+      , makeElement (Just "table") "hline" []
+      , makeElement (Just "table") "vline" []
+      , makeElement (Just "table") "header" [ ("children", Many TContent) ]
+      , makeElement (Just "table") "footer" [ ("children", Many TContent) ]
+      ],
+    makeElementWithScope Nothing
+      "grid"
+      [("children", Many TContent)]
+      [ makeElement (Just "grid") "cell" [ ("body", One TContent) ]
+      , makeElement (Just "grid") "hline" []
+      , makeElement (Just "grid") "vline" []
+      , makeElement (Just "grid") "header" [ ("children", Many TContent) ]
+      , makeElement (Just "grid") "footer" [ ("children", Many TContent) ]
+      ],
     makeElementWithScope
       Nothing
       "terms"
@@ -205,15 +229,11 @@ visualize =
 meta :: [(Identifier, Val)]
 meta =
   [ makeElement Nothing "bibliography" [("path", One (TString :|: TArray))],
-    makeElement
-      Nothing
-      "cite"
-      [ ("keys", Many TString),
-        ("supplement", One (TContent :|: TNone))
-      ],
+    makeElement Nothing "cite" [("key", One TLabel)],
     makeElement Nothing "document" [],
     makeElement Nothing "figure" [("body", One TContent)],
     makeElement Nothing "heading" [("body", One TContent)],
+    makeElement Nothing "quote" [("body", One TContent)],
     makeElement Nothing "layout" [("func", One TFunction)],
     makeElement
       Nothing
@@ -332,15 +352,7 @@ foundations =
     ( "type",
       makeFunction $ do
         (x :: Val) <- nthArg 1
-        pure $
-          VString $
-            case valType x of
-              TAlignment ->
-                case x of
-                  VAlignment (Just _) (Just _) -> "2d alignment"
-                  _ -> "alignment"
-              TDict -> "dictionary"
-              ty -> T.toLower . T.drop 1 . T.pack . show $ ty
+        pure $ VType $ valType x
     )
   ]
 
@@ -427,6 +439,11 @@ construct =
              pure $ VString $ T.pack [chr val] )
       ]
     ),
+    ( "version",
+        makeFunction $ do
+          xs <- asks positional >>= mapM fromVal
+          pure $ VVersion xs
+    ),
     ( "symbol",
       makeFunction $ do
         (t :: Text) <- nthArg 1
@@ -485,12 +502,14 @@ hexToRGB _ = fail "expected string"
 loadFileLazyBytes :: Monad m => FilePath -> MP m BL.ByteString
 loadFileLazyBytes fp = do
   operations <- evalOperations <$> getState
-  lift $ BL.fromStrict <$> loadBytes operations fp
+  root <- evalPackageRoot <$> getState
+  lift $ BL.fromStrict <$> loadBytes operations (root </> fp)
 
 loadFileText :: Monad m => FilePath -> MP m T.Text
 loadFileText fp = do
   operations <- evalOperations <$> getState
-  lift $ TE.decodeUtf8 <$> loadBytes operations fp
+  root <- evalPackageRoot <$> getState
+  lift $ TE.decodeUtf8 <$> loadBytes operations (root </> fp)
 
 getUTCTime :: Monad m => MP m UTCTime
 getUTCTime = (currentUTCTime . evalOperations <$> getState) >>= lift
@@ -556,7 +575,7 @@ dataLoading =
       makeFunction $ do
         fp <- nthArg 1
         t <- lift $ loadFileText fp
-        case Toml.decode (T.unpack t) of
+        case Toml.decode t of
           Toml.Failure e -> fail (unlines ("toml errors:" : e))
           Toml.Success _ v -> pure v
     ),
@@ -609,3 +628,28 @@ applyPureFunction (Function f) vals =
 initialEvalState :: MonadFail m => EvalState m
 initialEvalState =
   emptyEvalState { evalIdentifiers = [(BlockScope, standardModule)] }
+
+-- mDigitsRev, mDigits from the unmaintained digits package
+-- https://hackage.haskell.org/package/digits-0.3.1
+-- (c) 2009-2016 Henry Bucklow, Charlie Harvey -- BSD-3-Clause license.
+mDigitsRev :: Integral n
+    => n         -- ^ The base to use.
+    -> n         -- ^ The number to convert to digit form.
+    -> Maybe [n] -- ^ Nothing or Just the digits of the number in list form, in reverse.
+mDigitsRev base i = if base < 1
+                    then Nothing -- We do not support zero or negative bases
+                    else Just $ dr base i
+    where
+      dr _ 0 = []
+      dr b x = case base of
+                1 -> genericTake x $ repeat 1
+                _ -> let (rest, lastDigit) = quotRem x b
+                     in lastDigit : dr b rest
+
+-- | Returns the digits of a positive integer as a Maybe list.
+--   or Nothing if a zero or negative base is given
+mDigits :: Integral n
+    => n -- ^ The base to use.
+    -> n -- ^ The number to convert to digit form.
+    -> Maybe [n] -- ^ Nothing or Just the digits of the number in list form
+mDigits base i = reverse <$> mDigitsRev base i
