@@ -24,7 +24,13 @@ import Typst.Shorthands (mathSymbolShorthands)
 
 -- import Debug.Trace
 
-parseTypst :: FilePath -> Text -> Either ParseError [Markup]
+-- | Parse text into a list of 'Markup' (or a Parsec @ParseError@).
+parseTypst ::
+  -- | Filepath to Typst source text, only used for error messages
+  FilePath ->
+  -- | The Typst source text 
+  Text ->
+  Either ParseError [Markup]
 parseTypst fp inp =
   case runParser (spaces *> many pMarkup <* pEndOfContent) initialState fp inp of
     Left e -> Left e
@@ -60,8 +66,8 @@ ws = do
   inp <- getInput
   allowNewlines <- stAllowNewlines <$> getState
   let isSp c
-        | allowNewlines > 0 = c == ' ' || c == '\t' || c == '\n' || c == '\r'
-        | otherwise = c == ' ' || c == '\t'
+        | allowNewlines > 0 = isSpace c
+        | otherwise = isSpace c && c /= '\r' && c /= '\n'
   ( skipMany1 (void (satisfy isSp) <|> void pComment)
       *> updateState (\st -> st {stBeforeSpace = Just (p1, inp)})
     )
@@ -158,11 +164,11 @@ pEquation = do
 
 mathOperatorTable :: [[Operator Text PState Identity Markup]]
 mathOperatorTable =
-  [ -- precedence 6
-    [ Infix (attachBottom <$ op "_") AssocLeft,
-      Infix (attachTop <$ op "^") AssocLeft
+  [ -- precedence 7 -- attachment with number, e.g. a_1 (see #17)
+    [ Infix (attachBottom <$ (try (op "_" *> lookAhead mNumber))) AssocLeft,
+      Infix (attachTop <$ (try (op "^" *> lookAhead mNumber))) AssocLeft
     ],
-    -- precedence 5
+    -- precedence 6
     [ Postfix
         ( try $ do
             mbBeforeSpace <- stBeforeSpace <$> getState
@@ -173,6 +179,10 @@ mathOperatorTable =
             pure $ \expr -> MGroup Nothing Nothing [expr, args]
         )
     ],
+    -- precedence 5 -- attachment with non-number, e.g. a_x
+    [ Infix (attachBottom <$ op "_") AssocLeft,
+      Infix (attachTop <$ op "^") AssocLeft
+    ],
     -- precedence 4  -- factorial needs to take precedence over fraction
     [ Postfix (try $ do
                   mbBeforeSpace <- stBeforeSpace <$> getState
@@ -185,11 +195,17 @@ mathOperatorTable =
     ]
   ]
 
+
+ -- MAttach (Maybe bottom) (Maybe top) base
+
 attachBottom :: Markup -> Markup -> Markup
+attachBottom (MAttach Nothing y x) z = MAttach (Just (hideOuterParens z)) y x
+attachBottom z (MAttach Nothing y x) = MAttach (Just (hideOuterParens x)) y z
 attachBottom base x = MAttach (Just (hideOuterParens x)) Nothing base
 
 attachTop :: Markup -> Markup -> Markup
 attachTop (MAttach x Nothing y) z = MAttach x (Just (hideOuterParens z)) y
+attachTop z (MAttach x Nothing y) = MAttach x (Just (hideOuterParens y)) z
 attachTop base x = MAttach Nothing (Just (hideOuterParens x)) base
 
 makeFrac :: Markup -> Markup -> Markup
@@ -200,7 +216,16 @@ hideOuterParens (MGroup (Just "(") (Just ")") x) = MGroup Nothing Nothing x
 hideOuterParens x = x
 
 mathExpressionTable :: [[Operator Text PState Identity Expr]]
-mathExpressionTable = take 16 (cycle [[fieldAccess], [mathFunctionCall]])
+mathExpressionTable = take 16 (cycle [[mathFieldAccess], [mathFunctionCall]])
+
+mathFieldAccess :: Operator Text PState Identity Expr
+mathFieldAccess =
+  Postfix (FieldAccess <$> try (sym "." *> (Ident <$> pMathField)))
+ where
+  pMathField = lexeme $ do
+    d <- satisfy (\c -> isIdentStart c && c /= '_')
+    ds <- many $ satisfy (\c -> isIdentContinue c && c /= '_' && c /= '-')
+    pure $ Identifier $ T.pack (d : ds)
 
 mathFunctionCall :: Operator Text PState Identity Expr
 mathFunctionCall =
@@ -218,7 +243,12 @@ mExpr :: P Markup
 mExpr = Code <$> getPosition <*> pMathExpr
 
 pMathExpr :: P Expr
-pMathExpr = buildExpressionParser mathExpressionTable (pMathIdent <|> pLiteral)
+pMathExpr = buildExpressionParser mathExpressionTable
+               (pMathIdent <|> pMathLiteral)
+ where
+   pMathLiteral :: P Expr
+   pMathLiteral = Block . Content
+                    <$> many1 (mLiteral <|> mEscaped <|> mShorthand)
 
 pMathIdent :: P Expr
 pMathIdent =
@@ -237,20 +267,18 @@ pMathIdent =
 
 pMathIdentifier :: P Identifier
 pMathIdentifier = lexeme $ try $ do
-  c <- satisfy isIdentStart
-  cs <- many1 $ satisfy isMathIdentContinue
-  pure $ Identifier $ T.pack (c : cs)
-
-isMathIdentContinue :: Char -> Bool
-isMathIdentContinue c = isIdentContinue c && c /= '_' && c /= '-'
+  d <- satisfy (\c -> isIdentStart c && c /= '_')
+  ds <- many1 $ satisfy (\c -> isIdentContinue c && c /= '_' && c /= '-')
+  pure $ Identifier $ T.pack (d : ds)
 
 pMath :: P Markup
 pMath = buildExpressionParser mathOperatorTable pBaseMath
-  where
-    pBaseMath =
-      mNumber
+
+pBaseMath :: P Markup
+pBaseMath = mNumber
         <|> mLiteral
         <|> mEscaped
+        <|> mShorthand
         <|> mBreak
         <|> mAlignPoint
         <|> mExpr
@@ -260,11 +288,9 @@ pMath = buildExpressionParser mathOperatorTable pBaseMath
         <|> mSymbol
 
 mGroup :: P Markup
-mGroup =
-  mGrouped '(' ')' False
-    <|> mGrouped '{' '}' False
-    <|> mGrouped '[' ']' False
-    <|> mGrouped '|' '|' True
+mGroup = mGrouped '(' ')' False
+     <|> mGrouped '{' '}' False
+     <|> mGrouped '[' ']' False
 
 mGrouped :: Char -> Char -> Bool -> P Markup
 mGrouped op' cl requireMatch = withNewlines $ try $ do
@@ -358,28 +384,29 @@ mCode = lexeme $ char '#' *> (Code <$> getPosition <*> pBasicExpr)
 
 mMid :: P Markup
 mMid = try $ do
-  stBeforeSpace <$> getState >>= guard . isJust
+  getState >>= guard . isJust . stBeforeSpace
   void $ char '|' *> space *> ws
   pure $ MGroup Nothing Nothing [Nbsp, Text "|", Nbsp]
 
-mSymbol :: P Markup
-mSymbol =
-  getPosition >>= pSym
+mShorthand :: P Markup
+mShorthand =
+  getPosition >>= \pos ->
+   lexeme (Code pos <$> choice (map toShorthandParser shorthands))
  where
-  pSym pos = lexeme $
-    (Code pos <$> choice (map toShorthandParser shorthands))
-          <|> ( Text . T.singleton
-                  <$> satisfy (\c -> not (isSpace c) && c /= '$' && c /= '\\')
-              )
-  shorthands= reverse $ sortOn (T.length . fst) mathSymbolShorthands
+  shorthands = reverse (sortOn (T.length . fst) mathSymbolShorthands)
   toShorthandParser (short, symname) =
-    (toSym symname <$ try (string (T.unpack short)))
+    toSym symname <$ try (string (T.unpack short))
   toSym name =
     case map (Ident . Identifier) $ T.split (== '.') name of
       [] -> Literal None
       [i] -> i
       (i:is) -> foldr FieldAccess i is
 
+mSymbol :: P Markup
+mSymbol =
+  lexeme ( Text . T.singleton
+            <$> satisfy (\c -> not (isSpace c) && c /= '$' && c /= '\\'))
+
 withIndent :: Int -> P a -> P a
 withIndent indent pa = do
   oldIndent <- stIndent <$> getState
@@ -436,7 +463,7 @@ pLineComment :: P ()
 pLineComment = do
   void $ string "//"
   skipMany (satisfy (\c -> c /= '\n' && c /= '\r'))
-  void endOfLine
+  void endOfLine <|> eof
 
 pBlockComment :: P ()
 pBlockComment = do
@@ -485,7 +512,7 @@ pHardbreak =
 pBlankline :: P ()
 pBlankline = try $ do
   skipMany spaceChar
-  void (lookAhead (endOfLine)) <|> pEndOfContent
+  void (lookAhead endOfLine) <|> pEndOfContent
 
 pRawInline :: P Markup
 pRawInline =
@@ -558,12 +585,13 @@ pNonspaceWithBalancedBrackets parens brackets braces =
     <|> pure []
 
 pText :: P Markup
-pText =
-  Text . T.pack
-    <$> some
-      ( satisfy (\c -> not (isSpace c || isSpecial c))
-          <|> try ((char '*' <|> char '_') <* lookAhead alphaNum)
-      )
+pText = Text . mconcat <$> some
+  ((do xs <- some alphaNum
+       T.pack . (xs <>) <$>
+             try (some (char '*' <|> char '_') <* lookAhead alphaNum)
+        <|> pure (T.pack xs))
+ <|> (T.pack <$> some (satisfy (\c -> not (isSpace c || isSpecial c))))
+  )
 
 pEscaped :: P Markup
 pEscaped = Text . T.singleton <$> pEsc
@@ -622,7 +650,8 @@ pLabel =
   Label . T.pack
     <$> try
       ( char '<'
-          *> many1 (satisfy isIdentContinue <|> char '_' <|> char '.')
+          *> many1 (satisfy isIdentContinue <|>
+                    char '_' <|> char '.' <|> char ':')
           <* char '>'
       )
 
@@ -739,24 +768,18 @@ pBaseExpr =
   pLiteral
     <|> pKeywordExpr
     <|> pFuncExpr
-    <|> pBindExpr
     <|> pIdent
     <|> pArrayExpr
     <|> pDictExpr
     <|> inParens pExpr
-    <|> (Block . Content . (: []) <$> pEquation)
     <|> pLabel
+    <|> (Block . Content . (: [])
+         <$> lexeme (pRawBlock <|> pRawInline <|> pEquation))
     <|> pBlock
 
 pLiteral :: P Expr
 pLiteral =
-  Literal
-    <$> ( pNone
-            <|> pAuto
-            <|> pBoolean
-            <|> pNumeric
-            <|> pStr
-        )
+  Literal <$> ( pNone <|> pAuto <|> pBoolean <|> pNumeric <|> pStr )
 
 fieldAccess :: Operator Text PState Identity Expr
 fieldAccess = Postfix (FieldAccess <$> try (sym "." *> pIdent))
@@ -896,7 +919,7 @@ pNumeric = lexeme $ do
 pStr :: P Literal
 pStr = lexeme $ do
   void $ char '"'
-  String . T.pack <$> manyTill (pStrEsc <|> noneOf "\"\r\n") (char '"')
+  String . T.pack <$> manyTill (pStrEsc <|> satisfy (/= '"')) (char '"')
 
 pUnit :: P Unit
 pUnit =
@@ -967,18 +990,13 @@ pArrayExpr =
       )
         <|> (Array [] <$ optional (void $ sym ","))
 
--- dict-expr ::= '(' (':' | (pair (',' pair)* ','?)) ')'
+-- dict-expr ::= '(' (':' | ':'? (pair (',' pair)* ','?)) ')'
 -- pair ::= (ident | str) ':' expr
 pDictExpr :: P Expr
-pDictExpr = try $ inParens (pEmptyDict <|> pNonemptyDict)
+pDictExpr = try $ inParens (sym ":" *> (pNonemptyDict <|> pure (Dict mempty)) <|> pNonemptyDict)
   where
-    pEmptyDict = Dict mempty <$ sym ":"
     pNonemptyDict = Dict <$> sepEndBy1 (pSpread <|> pPair) (sym ",")
-    pPair = Reg <$> ((,) <$> pKey <*> try (sym ":" *> pExpr))
-    pKey = pIdentifier <|> pStrKey
-    pStrKey = do
-      String t <- pStr
-      pure $ Identifier t
+    pPair = Reg <$> ((,) <$> pExpr <*> try (sym ":" *> pExpr))
 
 pSpread :: P (Spreadable a)
 pSpread = try $ string ".." *> (Spr <$> pExpr)
@@ -1028,9 +1046,7 @@ pArg :: P Arg
 pArg = pKeyValArg <|> pSpreadArg <|> pNormalArg
   where
     pKeyValArg = KeyValArg <$> try (pIdentifier <* sym ":") <*> pExpr
-    pNormalArg =
-      NormalArg
-        <$> ((Block . Content . (: []) <$> lexeme (pRawBlock <|> pRawInline)) <|> pExpr)
+    pNormalArg = NormalArg <$> pExpr
     pSpreadArg = SpreadArg <$> try (string ".." *> pExpr)
 
 -- params ::= '(' (param (',' param)* ','?)? ')'
@@ -1108,7 +1124,11 @@ pLetExpr = do
 -- set-expr ::= 'set' expr args
 pSetExpr :: P Expr
 pSetExpr = do
+  oldAllowNewlines <- stAllowNewlines <$> getState
+  -- see #23 -- 'set' doesn't go with 'if' unless it's on the same line
+  updateState $ \st -> st {stAllowNewlines = 0}
   set <- pKeyword "set" *> (Set <$> pQualifiedIdentifier <*> pArgs)
+  updateState $ \st -> st {stAllowNewlines = oldAllowNewlines}
   addCondition <- option id $ pKeyword "if" *> ((\c x -> If [(c, x)]) <$> pExpr)
   pure $ addCondition set
 
@@ -1145,11 +1165,17 @@ pImportExpr :: P Expr
 pImportExpr = pKeyword "import" *> (Import <$> pExpr <*> pImportItems)
   where
     pImportItems =
-      option NoIdentifiers $
-        sym ":"
+        (sym ":"
           *> ( (AllIdentifiers <$ sym "*")
-                 <|> (SomeIdentifiers <$> sepEndBy pIdentifier (sym ","))
+                 <|> (SomeIdentifiers
+                       <$> sepEndBy pIdentifierAs (sym ","))
              )
+        ) <|> (NoIdentifiers <$> pAs)
+    pIdentifierAs = do
+      ident <- pIdentifier
+      mbAs <- pAs
+      pure (ident, mbAs)
+    pAs = option Nothing $ Just <$> (pKeyword "as" *> pIdentifier)
 
 pBreakExpr :: P Expr
 pBreakExpr = Break <$ pKeyword "break"
@@ -1168,7 +1194,3 @@ pReturnExpr = do
 
 pIncludeExpr :: P Expr
 pIncludeExpr = Include <$> (pKeyword "include" *> pExpr)
-
-pBindExpr :: P Expr
-pBindExpr =
-  Binding <$> try (pBind <* lookAhead (op "="))
