@@ -45,6 +45,7 @@ module Typst.Types
     joinVals,
     prettyVal,
     valToContent,
+    prettyType,
     repr,
     Attempt (..),
   )
@@ -71,7 +72,7 @@ import Data.Vector (Vector)
 import qualified Data.Vector as V
 import Text.Parsec
 import qualified Toml
-import qualified Toml.FromValue as Toml
+import qualified Toml.Schema as Toml
 import qualified Toml.Pretty as Toml
 import qualified Text.PrettyPrint as P
 import Text.Read (readMaybe)
@@ -81,36 +82,69 @@ import Data.Time (UTCTime, Day, DiffTime, timeOfDayToTime, localDay, localTimeOf
 import Data.Time.Format (defaultTimeLocale, formatTime)
 import System.Directory (XdgDirectory(..))
 
+-- | A Typst value. More documentation can be found in the 
+-- [Foundations chapter](https://typst.app/docs/reference/foundations/)
+-- of the Typst reference manual. A more concise (but somewhat outdated) 
+-- summary can also be found in 
+-- [L. Mädje "Typst: a programmable markup language for typesetting", page 32-33](https://www.user.tu-berlin.de/laurmaedje/programmable-markup-language-for-typesetting.pdf).
 data Val
+  -- | The @none@ value, indicates the absence of any other value.
   = VNone
+  -- | The @auto@ value, used to automatically set an appropriate value.
   | VAuto
+  -- | A @bool@ value.
   | VBoolean !Bool
+  -- | An @int@ value.
   | VInteger !Integer
+  -- | A @float@ value.
   | VFloat !Double
+  -- | A @ratio@ value, a proportion of a certain whole, for example @50%@.
   | VRatio !Rational
+  -- | A @length@ or a @relative@ value.
   | VLength !Length
+  -- | An @alignment@ value, indicating the alignment of some content along both
+  -- the horizontal and vertical axes.
   | VAlignment (Maybe Horiz) (Maybe Vert)
-  | VAngle !Double -- degrees
+  -- | An @angle@ value (expressed internally in degrees).
+  | VAngle !Double
+  -- | A @fraction@ value, defining the proportions of remaing space is 
+  -- to be distributed, e.g. @2 fr@.
   | VFraction !Double
+  -- | A @color@ value. Not all Typst color spaces are supported; 
+  -- only @rgb@, @cmyk@, and @luma@ are available. 
+  -- See issue [#35](https://github.com/jgm/typst-hs/issues/35#issuecomment-1926182040).
   | VColor !Color
+  -- | A @symbol@ value, representing a Unicode symbol.
   | VSymbol !Symbol
+  -- | A UTF-8 encoded text @string@.
   | VString !Text
+  -- | A @regex@ (regular expression). See 'RE' for details.
   | VRegex !RE
+  -- | A @datetime@ value, a date, a time, or a combination of both.
   | VDateTime (Maybe Day) (Maybe DiffTime)
+  -- | A @content@ value, see 'Content' for more details.
   | VContent (Seq Content)
+  -- | An @array@ value, for example @(10, 20, 30)@.
   | VArray (Vector Val)
+  -- | A @dictionary@ value, for example @(a:20, b:30)@.
   | VDict (OM.OMap Identifier Val)
   | VTermItem (Seq Content) (Seq Content)
+  -- | A @direction@ to lay out content.
   | VDirection Direction
+  -- | A Typst function.
   | VFunction (Maybe Identifier) (M.Map Identifier Val) Function
   | -- first param is Just ident if element function
     -- second param is a map of subfunctions in this function's scope
+    -- | Positional and named function arguments
     VArguments Arguments
+  -- | A @label@ to some element, for example @<hello>@. 
   | VLabel !Text
   | VCounter !Counter
   | VSelector !Selector
   | VModule Identifier (M.Map Identifier Val)
   | VStyles -- just a placeholder for now
+  | VVersion [Integer]
+  | VType !ValType
   deriving (Show, Eq, Typeable)
 
 instance FromJSON Val where
@@ -126,19 +160,20 @@ instance FromJSON Val where
 instance Toml.FromValue Val where
   fromValue = pure . tomlToVal
 
-tomlToVal :: Toml.Value -> Val
-tomlToVal (Toml.Bool x) = VBoolean x
-tomlToVal (Toml.Integer x) = VInteger x
-tomlToVal (Toml.String x) = VString (T.pack x)
-tomlToVal (Toml.Float x) = VFloat x
-tomlToVal (Toml.TimeOfDay x) = VDateTime Nothing (Just (timeOfDayToTime x))
-tomlToVal (Toml.Day x) = VDateTime (Just x) Nothing
-tomlToVal (Toml.LocalTime x) = VDateTime (Just (localDay x)) (Just (timeOfDayToTime (localTimeOfDay x)))
-tomlToVal (Toml.Array x) = VArray (V.fromList (map tomlToVal x))
-tomlToVal (Toml.Table x) = VDict (OM.fromList [(Identifier (T.pack k), tomlToVal v) | (k,v) <- M.assocs x])
+tomlToVal :: Toml.Value' a -> Val
+tomlToVal (Toml.Bool' _ x) = VBoolean x
+tomlToVal (Toml.Integer' _ x) = VInteger x
+tomlToVal (Toml.Text' _ x) = VString x
+tomlToVal (Toml.Double' _ x) = VFloat x
+tomlToVal (Toml.TimeOfDay' _ x) = VDateTime Nothing (Just (timeOfDayToTime x))
+tomlToVal (Toml.Day' _ x) = VDateTime (Just x) Nothing
+tomlToVal (Toml.LocalTime' _ x) = VDateTime (Just (localDay x)) (Just (timeOfDayToTime (localTimeOfDay x)))
+tomlToVal (Toml.List' _ x) = VArray (V.fromList (map tomlToVal x))
+tomlToVal (Toml.Table' _ (Toml.MkTable x)) = VDict (OM.fromList [(Identifier k, tomlToVal v) | (k,(_,v)) <- M.assocs x])
   -- typst specifies that unsupported datetimes map to strings and we don't have a place for the timezone
-tomlToVal v@Toml.ZonedTime{} = VString (T.pack (show (Toml.prettyValue v)))
+tomlToVal v@Toml.ZonedTime'{} = VString (T.pack (show (Toml.prettyValue v)))
 
+-- | A Typst type, see documentation for 'Val'.
 data ValType
   = TNone
   | TAuto
@@ -168,9 +203,11 @@ data ValType
   | TLabel
   | TCounter
   | TLocation
+  | TVersion
+  | TType
   | TAny
   | ValType :|: ValType
-  deriving (Show, Eq, Typeable)
+  deriving (Show, Eq, Ord, Typeable)
 
 valType :: Val -> ValType
 valType v =
@@ -202,6 +239,8 @@ valType v =
     VModule {} -> TModule
     VSelector {} -> TSelector
     VStyles {} -> TStyles
+    VVersion {} -> TVersion
+    VType {} -> TType
 
 hasType :: ValType -> Val -> Bool
 hasType TAny _ = True
@@ -367,6 +406,16 @@ instance Compare Val where
   comp (VDict m1) (VDict m2) =
     Just $ liftCompare (\x y -> fromMaybe LT (comp x y)) (OM.toMap m1) (OM.toMap m2)
   comp (VFunction (Just i1) _ _) (VFunction (Just i2) _ _) = Just $ compare i1 i2
+  comp (VType ty1) (VType ty2) = Just $ compare ty1 ty2
+  comp (VType TInteger) (VString "integer") = Just EQ
+  comp (VString "integer") (VType TInteger) = Just EQ
+  comp (VType ty) (VString s) = Just $ compare (prettyType ty) s
+  comp (VString s) (VType ty)  = Just $ compare s (prettyType ty)
+  comp (VVersion as) (VVersion bs)
+    | length as > length bs =
+       Just $ compare as (bs ++ replicate (length as - length bs) 0)
+    | otherwise =
+       Just $ compare (as ++ replicate (length as - length bs) 0) bs
   comp _ _ = Nothing
 
 instance Ord Val where
@@ -751,7 +800,11 @@ data Color
   | Luma Rational
   deriving (Show, Eq, Ord, Typeable)
 
-data Direction = Ltr | Rtl | Ttb | Btt
+data Direction 
+  = Ltr -- ^ Left to right
+  | Rtl -- ^ Right to left
+  | Ttb -- ^ Top to bottom
+  | Btt -- ^ Bottom to top
   deriving (Show, Eq, Ord, Typeable)
 
 prettyVal :: Val -> P.Doc
@@ -802,7 +855,7 @@ prettyVal expr =
         )
     VDirection d -> text $ T.toLower $ T.pack $ show d
     VFunction _ _ _ -> mempty
-    VLabel _ -> mempty
+    VLabel t -> text $ "<" <> t <> ">"
     VCounter _ -> mempty
     VColor (RGB r g b o) ->
       "rgb("
@@ -852,6 +905,13 @@ prettyVal expr =
     VSymbol (Symbol t _ _) -> text t
     VSelector _ -> mempty
     VStyles -> mempty
+    VVersion xs -> text $ T.intercalate "." (map (T.pack . show) xs)
+    VType ty -> text $ prettyType ty
+
+prettyType :: ValType -> Text
+prettyType TDict = "dictionary"
+prettyType TInteger = "int"
+prettyType x = T.toLower . T.pack . drop 1 . show $ x
 
 escString :: Text -> P.Doc
 escString =
